#include <stdlib.h>
#include "stack.h"
#include "strerr.h"
#include "sml.h"
#include "timeout.h"
#include "buffer.h"
#include "context.h"
#include "fmt.h"

#define NumCells 5000

ContextStack *bottom;
static ContextStack *nextfree;

void
ContextStack_init(void)
{
    bottom = (ContextStack *) malloc(NumCells * sizeof *bottom);
    if (!bottom) quit();

    bottom->next = 0;
    bottom->x.flags = 0;
    bottom->x.size = 0;
    bottom->x.color = 0;
    bottom->x.tag = 0;

    nextfree = bottom + 1;
}

static ContextStack *
getcell(void)
{
    if (nextfree == bottom + NumCells) quit();
    return nextfree++;
}

ContextStack *
ContextStack_push(ContextStack *top, TagTypes tag)
{
    ContextStack *new_top = getcell();

    new_top->next = top;
    new_top->x = top->x;
    new_top->x.tag = tag;

    switch (tag) {
	case Tag_PL: new_top->x.flags = 0; break;
	case Tag_U: new_top->x.flags++; break;
	case Tag_B: new_top->x.flags |= B; break;
	case Tag_I: new_top->x.flags |= I; break;
	case Tag_TT: new_top->x.flags |= TT; break;
	case Tag_S: new_top->x.flags &= ~EM; new_top->x.flags |= S; break;
	case Tag_EM: new_top->x.flags |= EM; break;
	case Tag_0: case Tag_1: case Tag_2: case Tag_3: case Tag_4:
	case Tag_5: case Tag_6: case Tag_7: case Tag_8: case Tag_9:
	    new_top->x.size = '0' + tag - Tag_0; break;
	case Tag_r: new_top->x.color = 'r'; break;
	case Tag_g: new_top->x.color = 'g'; break;
	case Tag_b: new_top->x.color = 'b'; break;
	case Tag_w: new_top->x.color = 'w'; break;
	case Tag_c: new_top->x.color = 'c'; break;
	case Tag_m: new_top->x.color = 'm'; break;
	case Tag_y: new_top->x.color = 'y'; break;
	case Tag_k: new_top->x.color = 'k'; break;
	case Tag_Count: /* shut up gcc */ break;
    }

    return new_top;
}

static ContextStack *
gc_recurse(ContextStack *mark, ContextStack *top)
{
    ContextStack *new_top;

    if (!top || top < mark) return top;

    top->next = gc_recurse(mark, top->next);
    new_top = getcell();
    *new_top = *top;
    return new_top;
}

ContextStack *
ContextStack_gc(ContextStack *mark, ContextStack *top)
{
    nextfree = mark;
    return gc_recurse(mark, top);
}

ContextStack *
ContextStack_mark(void)
{
    return nextfree;
}

void
ContextStack_dump(char *message, int lookahead, ContextStack *top)
{
    char b[FMT_ULONG];

    buffer_puts(buffer_2, message);
    buffer_put(buffer_2, b, fmt_ulong(b, lookahead));
    buffer_put(buffer_2, " ", 1);

    while (1) {
	Context_dump(top->x);
	if (!top->next) break;
	buffer_puts(buffer_2, tag_close[(int) top->x.tag]);
	top = top->next;
    }

    buffer_putflush(buffer_2, "\n", 1);
}

